# -*- coding: utf-8 -*-
#
# Copyright (c) 2019-2022, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the BSD 3-Clause License:
# http://opensource.org/licenses/BSD-3-Clause
#


from __future__ import annotations


import asyncio
import atexit
import collections
import json
import pathlib
import sys
import time
import zipfile
from typing import Callable, Dict, List, Optional, Set
from .. import err
from .. import message
from .. import util
from ..code_chunks import CodeChunk, CodeKey
from ..code_collections import Session, Source
from ..codebraid_defaults import CodebraidDefaults
from ..progress import Progress
from ..version import __version__ as codebraid_version
from . import exec_builtin
from . import exec_jupyter




class CodeProcessor(object):
    '''
    Process code chunks.  This can involve executing code, copying code and/or
    output between code chunks, and creating source files.
    '''
    def __init__(self,
                 *,
                 code_chunks: List[CodeChunk],
                 cross_origin_sessions: bool,
                 no_cache: bool,
                 cache_path: pathlib.Path,
                 cache_key: str,
                 origin_paths_for_cache: Optional[List[pathlib.Path]],
                 codebraid_defaults: CodebraidDefaults,
                 only_code_output: bool,
                 progress: Progress):
        self.code_chunks = code_chunks
        self.cross_origin_sessions = cross_origin_sessions
        self.no_cache = no_cache
        self.cache_path = cache_path
        self.cache_key = cache_key
        self.origin_paths_for_cache = origin_paths_for_cache
        self._origin_paths_for_cache_as_strings: Optional[List[str]]
        if origin_paths_for_cache is None:
            self._origin_paths_for_cache_as_strings = None
        else:
            self._origin_paths_for_cache_as_strings = [p.as_posix() for p in origin_paths_for_cache]
        self.codebraid_defaults = codebraid_defaults
        self._only_code_output = only_code_output
        self._progress = progress

        self._old_cache_index: Optional[Dict] = None
        self._cache_key_path = cache_path / cache_key
        self._cache_index_path = cache_path / cache_key / f'{cache_key}_index.zip'
        self._cache_lock_path = cache_path / cache_key / f'{cache_key}.lock'

        self._sessions: Dict[CodeKey, Session] = util.KeyDefaultDict(
            lambda x: Session(x, codebraid_defaults=self.codebraid_defaults)
        )
        self._session_hash_root_sets: Dict[str, Set] = collections.defaultdict(set)
        self._cached_sessions: Set[Session] = set()
        self._named_code_chunks: Dict[str, CodeChunk] = {}

        self._sources: Dict[CodeKey, Source] = util.KeyDefaultDict(lambda x: Source(x, codebraid_defaults=self.codebraid_defaults))

        # Use `atexit` to improve the odds of cleanup in the event of an
        # unexpected exit.  `cleanup()` ends by invoking
        # `atexit.unregister(self.cleanup)`.
        atexit.register(self.cleanup)


    @property
    def exit_code(self) -> int:
        code = 0b00000000
        if any(s.status.prevent_exec for s in self._sessions.values()):
            code ^= 0b00000100
        if (any(s.status.has_errors and not s.status.prevent_exec for s in self._sessions.values()) or
                any(s.status.has_errors for s in self._sources.values())):
            code ^= 0b00001000
        if any(s.status.has_warnings for s in self._sessions.values()):
            code ^= 0b00010000
        # Once there are warnings related to document build, add condition:
        #   code ^= 0b0010000
        return code


    def process(self):
        '''
        Set additional code chunk attributes that are inconvenient or
        impossible to determine until all code chunks are assembled.  Group
        code chunks into sessions and sources.
        '''
        self._index_named_code_chunks()
        self._resolve_code_copying()
        self._prep_cache()
        self._create_sessions_and_sources()
        if self._only_code_output and self._cached_sessions:
            # Resolve output copying as early as possible
            self._resolve_output_copying(from_cache=True)


    def exec(self):
        '''
        Execute code, update code, and copy code output.
        '''
        builtin_sessions = []
        jupyter_sessions = []
        for session in self._sessions.values():
            if not session.needs_exec:
                continue
            if session.jupyter_kernel is None:
                builtin_sessions.append(session)
            else:
                jupyter_sessions.append(session)
        # Use `atexit` to improve the odds of updating the cache index in the
        # event of an unexpected exit.  Any session caches that are
        # successfully updated can be used in the future, regardless of
        # whether an index is created.  However, if an index is not created,
        # there is no guarantee that unused cache files will be deleted.
        if builtin_sessions or jupyter_sessions:
            atexit.register(self._update_cache_index)
            if sys.platform == 'win32':
                # For Windows, need different event loops depending on whether
                # code execution uses the built-in system (asyncio subprocess)
                # or a Jupyter kernel:  https://bugs.python.org/issue37373.
                original_loop_policy = asyncio.get_event_loop_policy()
                if builtin_sessions:
                    if not isinstance(original_loop_policy, asyncio.windows_events.WindowsProactorEventLoopPolicy):
                        asyncio.set_event_loop_policy(asyncio.WindowsProactorEventLoopPolicy())
                    asyncio.run(self._exec_sessions(builtin_sessions=builtin_sessions))
                    asyncio.set_event_loop_policy(original_loop_policy)
                if jupyter_sessions:
                    if not isinstance(original_loop_policy, asyncio.windows_events.WindowsSelectorEventLoopPolicy):
                        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
                    asyncio.run(self._exec_sessions(jupyter_sessions=jupyter_sessions))
                    asyncio.set_event_loop_policy(original_loop_policy)
            else:
                asyncio.run(self._exec_sessions(builtin_sessions=builtin_sessions, jupyter_sessions=jupyter_sessions))
            self._update_cache_index()
            atexit.unregister(self._update_cache_index)
        self._resolve_output_copying()

    async def _exec_sessions(self, *,
                             builtin_sessions: Optional[List[Session]]=None,
                             jupyter_sessions: Optional[List[Session]]=None):
        ticktock = asyncio.create_task(self._progress.ticktock())
        max_concurrent_jobs = asyncio.Semaphore(1)
        coroutines = []
        if builtin_sessions is not None:
            coroutines.extend(
                [self._exec_session(exec_builtin.exec, session, max_concurrent_jobs) for session in builtin_sessions]
            )
        if jupyter_sessions is not None:
            coroutines.extend(
                [self._exec_session(exec_jupyter.exec, session, max_concurrent_jobs) for session in jupyter_sessions]
            )
        await asyncio.gather(*coroutines)
        ticktock.cancel()


    async def _exec_session(self, exec_func: Callable, session: Session, max_concurrent_jobs: asyncio.Semaphore):
        async with max_concurrent_jobs:
            await exec_func(session, cache_key_path=self._cache_key_path, progress=self._progress)
        self._update_session_cache(session)


    def _index_named_code_chunks(self):
        '''
        Index named code chunks, and mark code chunks with duplicate names.
        '''
        named_code_chunks = self._named_code_chunks
        for cc in self.code_chunks:
            cc_name = cc.options.get('name')
            if cc_name is not None:
                if cc_name not in named_code_chunks:
                    named_code_chunks[cc_name] = cc
                    continue
                other_cc = named_code_chunks[cc_name]
                src = other_cc.origin_name
                lineno = other_cc.origin_start_line_number
                msg = (f'Code chunk names must be unique; "{cc_name}" is already used in "{src}" near line {lineno}')
                cc.errors.append(message.SourceError(msg))


    def _resolve_code_copying(self):
        '''
        For code chunks with copying, handle the code copying.  The output
        copying for commands like "paste" must be handled separately later,
        after code is executed.
        '''
        named_chunks = self._named_code_chunks
        unresolved_chunks = []
        for cc in self.code_chunks:
            copy_chunk_names = cc.options.get('copy')
            if copy_chunk_names is not None:
                copy_chunks = [named_chunks.get(name) for name in copy_chunk_names]
                if any(x is None for x in copy_chunks):
                    unknown_names = ', '.join(f'"{name}"'
                                              for name, copied_cc in zip(copy_chunk_names, copy_chunks)
                                              if copied_cc is None)
                    cc.errors.append(message.SourceError(f'Unknown name(s) for copying: {unknown_names}'))
                    continue
                cc.copy_chunks.extend(copy_chunks)
                unresolved_chunks.append(cc)
        still_unresolved_chunks = []
        while True:
            for cc in unresolved_chunks:
                if any(copied_cc.errors.prevent_exec for copied_cc in cc.copy_chunks):
                    traceback_list = []
                    for copied_cc in cc.copy_chunks:
                        if copied_cc.errors.prevent_exec:
                            name = copied_cc.options['name']
                            src = copied_cc.origin_name
                            lineno = copied_cc.origin_start_line_number
                            traceback_list.append(f'  * "{name}" ("{src}" near line {lineno})')
                    traceback = '\n'.join(traceback_list)
                    msg = f'Code chunk(s) have error(s) that prevent copying:\n{traceback}'
                    cc.errors.append(message.SourceError(msg))
                elif any(not copied_cc.code_lines for copied_cc in cc.copy_chunks):
                    still_unresolved_chunks.append(cc)
                else:
                    cc.copy_code()
            if not still_unresolved_chunks:
                break
            if len(still_unresolved_chunks) == len(unresolved_chunks):
                # Locate circular copying dependencies.  The case of a code
                # chunk trying to copy itself directly is already handled
                # during code chunk creation.
                for cc in still_unresolved_chunks:
                    copy_path_list = [cc]
                    copy_path_set = set(copy_path_list)
                    copy_state = [cc.copy_chunks.copy()]
                    while copy_state:
                        try:
                            last_cc = copy_state[-1].pop()
                        except IndexError:
                            copy_state.pop()
                            continue
                        while len(copy_path_list) > len(copy_state):
                            copy_path_set.remove(copy_path_list.pop())
                        copy_path_list.append(last_cc)
                        if last_cc in copy_path_set:
                            start_circular_name = copy_path_list[1].options['name']
                            traceback_list = []
                            for copied_cc in copy_path_list[1:]:
                                name = copied_cc.options['name']
                                src = copied_cc.origin_name
                                lineno = copied_cc.origin_start_line_number
                                if copied_cc is last_cc:
                                    traceback_list.append(f' => "{name}" ("{src}" near line {lineno})')
                                else:
                                    traceback_list.append(f' -> "{name}" ("{src}" near line {lineno})')
                            traceback = '\n'.join(traceback_list)
                            msg = f'Circular dependency in copying "{start_circular_name}":\n{traceback}'
                            cc.errors.append(message.SourceError(msg))
                            break
                        copy_path_set.add(last_cc)
                        if 'copy' in last_cc.options:
                            copy_state.append(last_cc.copy_chunks.copy())
                still_unresolved_chunks = [cc for cc in still_unresolved_chunks if not cc.errors.prevent_exec]
            unresolved_chunks = still_unresolved_chunks
            still_unresolved_chunks = []

    def _resolve_output_copying(self, *, from_cache: bool=False):
        '''
        For code chunks with copying, handle the output copying.  The output
        copying for commands like "paste" must be handled after code is
        executed, while any code copying must be handled before so that it is
        available for execution.
        '''
        # There is no need to check for undefined names or circular
        # dependencies; that's already been handled in
        # `_resolve_code_copying()`.  Still need to check for errors that
        # prevent copying, since there can be runtime source errors or other
        # errors related to output.  Code chunks with
        # `.needs_to_copy == False` are copying code-only code chunks and thus
        # have no output for copying.
        unresolved_chunks = [cc for cc in self.code_chunks
                             if cc.command == 'paste' and not cc.errors.prevent_exec and cc.needs_to_copy]
        still_unresolved_chunks = []
        while True:
            for cc in unresolved_chunks:
                if any(copied_cc.errors.prevent_exec for copied_cc in cc.copy_chunks):
                    traceback_list = []
                    for copied_cc in cc.copy_chunks:
                        if copied_cc.errors.prevent_exec:
                            name = copied_cc.options['name']
                            src = copied_cc.origin_name
                            lineno = copied_cc.origin_start_line_number
                            traceback_list.append(f'  * "{name}" ("{src}" near line {lineno})')
                    traceback = '\n'.join(traceback_list)
                    msg = f'Code chunk(s) have error(s) that prevent copying:\n{traceback}'
                    cc.errors.append(message.SourceError(msg))
                    self._progress.source_chunk_complete(cc.source, chunk=cc)
                    if all(not cc.needs_to_copy or cc.errors.prevent_exec for cc in cc.source.code_chunks):
                        self._progress.source_finished(cc.source)
                elif any(copied_cc.needs_to_copy for copied_cc in cc.copy_chunks):
                    still_unresolved_chunks.append(cc)
                else:
                    cc.copy_output()
                    self._progress.source_chunk_complete(cc.source, chunk=cc)
                    if all(not cc.needs_to_copy or cc.errors.prevent_exec for cc in cc.source.code_chunks):
                        self._progress.source_finished(cc.source)
            if not still_unresolved_chunks or (from_cache and len(unresolved_chunks) == len(still_unresolved_chunks)):
                break
            unresolved_chunks = still_unresolved_chunks
            still_unresolved_chunks = []


    def _create_sessions_and_sources(self):
        '''
        Assemble code chunks into sessions and sources.
        '''
        placeholder_lang_num = 0
        for cc in self.code_chunks:
            if cc.execute or 'session' in cc.options:
                # `cc.execute` covers valid commands, 'session' key covers
                # invalid commands with named sessions
                code_collection_name = cc.options['session']
                code_collection_type = 'session'
            else:
                code_collection_name = cc.options['source']
                code_collection_type = 'source'
            if self.cross_origin_sessions:
                origin_name = None
            else:
                origin_name = cc.origin_name
            cc.key = CodeKey(cc.options['lang'], code_collection_name, code_collection_type, origin_name)
            if cc.options['lang'] is None or cc.options['inherited_lang']:
                cc.options['placeholder_lang'] = f'{placeholder_lang_num}'
                placeholder_lang_num += 1
            if cc.execute:
                self._sessions[cc.key].append(cc)
            else:
                self._sources[cc.key].append(cc)
        for session in self._sessions.values():
            session.finalize()
            if session.status.prevent_exec:
                session.needs_exec = False
            else:
                self._session_hash_root_sets[session.hash_root].add(session)
                if self._load_session_cache(session):
                    session.needs_exec = False
        for source in self._sources.values():
            source.finalize()

        # Need to register collections before reporting any progress
        self._progress.register_code_collections(sessions=self._sessions.values(), sources=self._sources.values())
        for session in self._sessions.values():
            if session.status.prevent_exec:
                self._progress.session_errors_prevent_exec(session)
                for cc in session.code_chunks:
                    self._progress.session_chunk_complete_no_exec(session, chunk=cc)
                self._progress.session_finished(session)
            elif session in self._cached_sessions:
                self._progress.session_load_cache(session)
                for cc in session.code_chunks:
                    self._progress.session_chunk_complete_no_exec(session, chunk=cc)
                self._progress.session_finished(session)
        for source in self._sources.values():
            for cc in source.code_chunks:
                if not cc.needs_to_copy or cc.errors.prevent_exec:
                    self._progress.source_chunk_complete(source, chunk=cc)
            if all(not cc.needs_to_copy or cc.errors.prevent_exec for cc in source.code_chunks):
                self._progress.source_finished(source)


    def _prep_cache(self):
        '''
        Prepare the cache.

        A document's cache is located at `<cache_path>/<cache_key>/`.  If the
        document is read from stdin, <cache_key> is a constant.  If the
        document is read from file(s), <cache_key> is derived from a hash of
        the absolute source path(s).  Thus, the cache typically depends on
        source locations.  To provide some cross-device, cross-platform cache
        compatibility, source paths always use `~` to represent the user's
        home directory, even under Windows.

        Each session has a corresponding file in the document cache.  This
        file contains cacheable errors, cacheable warnings, and cacheable
        output (primarily text-based) such as stdout, sterr, and rich output.
        This data is stored in compressed JSON format.  For rich output,
        additional files such as images are also created.  While each session
        typically has its own separate file in the document cache, sessions
        can share a file in the rare case that their identifying hashes have
        identical starting sequences (`.hash_root`).

        The file `<cache_key>_index.zip` is used for managing the cache.  It
        includes all document source path(s) and a complete list of all cache
        files created (including itself).

        The cache is compatible with multiple documents being built
        simultaneously within a single `<cache_path>`, since the cache for
        each build will be located in its own subdirectory.  However, for a
        given document, only one build at a time is possible.  A lock file
        `<cache_path>/<cache_key>.lock` is used to enforce this.

        Notice that in cleaning outdated or invalid cache material, there is
        never wholesale directory removal.  Only files created directly by
        Codebraid are deleted.  Users should never manually create files in
        the cache, but if that happens those files should not be deleted
        unless they overwrite existing cache files.
        '''
        # Even when `no_cache == True`, need cache dir to store temp files
        # during build, such as rich output images
        self._cache_key_path.mkdir(parents=True, exist_ok=True)
        max_lock_wait = 5
        lock_check_interval = 0.1
        lock_time = 0
        while True:
            try:
                self._cache_lock_path.touch(exist_ok=False)
            except FileExistsError:
                if lock_time > max_lock_wait:
                    raise err.CodebraidError(
                        f'The cache for this document is locked.  This can happen if multiple processes attempt to '
                        f'build the document at the same time.  If Codebraid was forced to exit or encountered a '
                        f'severe error during the last run, it is also possible that the cache was corrupted.  In '
                        f'that case, the cache lock file "{self._cache_lock_path}" may need to be deleted manually.')
                time.sleep(lock_check_interval)
                lock_time += lock_check_interval
            else:
                break

        try:
            with zipfile.ZipFile(str(self._cache_index_path)) as zf:
                with zf.open('index.json') as f:
                    cache_index = json.load(f)
        except (FileNotFoundError, KeyError, json.JSONDecodeError):
            pass
        else:
            # These checks include handling source hash collisions and
            # corrupted caches.  The checks assume that if the key
            # 'codebraid_version' is present, then this is a valid Codebraid
            # index.
            if (self.no_cache or
                    not isinstance(cache_index, dict) or
                    cache_index.get('codebraid_version') != codebraid_version or
                    cache_index['origins'] != self._origin_paths_for_cache_as_strings or
                    not all((self._cache_key_path / f).is_file() for f in cache_index['files'])):
                if isinstance(cache_index, dict) and 'codebraid_version' in cache_index:
                    for f in cache_index['files']:
                        try:
                            (self._cache_key_path / f).unlink()
                        except FileNotFoundError:
                            pass
                else:
                    try:
                        self._cache_index_path.unlink()
                    except FileNotFoundError:
                        pass
            else:
                self._old_cache_index = cache_index


    def _load_session_cache(self, session: Session) -> bool:
        '''
        Load cached output for a session, if it exists.

        There's no need to check that all cache files are present in the case
        of rich output that results in additional files, because that is done
        during cache prep when the cache index is loaded.
        '''
        if self.no_cache or session.status.prevent_exec:
            return False
        if session in self._cached_sessions:
            return True
        session_cache_path = self._cache_key_path / f'{session.hash_root}.zip'
        try:
            with zipfile.ZipFile(str(session_cache_path)) as zf:
                with zf.open('cache.json') as f:
                    saved_cache = json.load(f)
        except (FileNotFoundError, KeyError, json.decoder.JSONDecodeError):
            return False
        if (not isinstance(saved_cache, dict) or
                'codebraid_version' not in saved_cache or
                saved_cache['codebraid_version'] != codebraid_version):
            return False
        try:
            session_cache = saved_cache['cache'][session.hash]
        except KeyError:
            return False
        for msg_name, msg_dict in session_cache['session_errors']:
            session.errors.append(message.message_name_to_class_map[msg_name](**msg_dict))
        for msg_name, msg_dict in session_cache['session_warnings']:
            session.warnings.append(message.message_name_to_class_map[msg_name](**msg_dict))
        session.files = session_cache['session_files']
        for index, chunk_cache in session_cache['code_chunks'].items():
            # `int(index)` because keys from JSON cache are strings
            chunk = session.code_chunks[int(index)]
            errors = chunk_cache.pop('errors', None)
            if errors is not None:
                for msg_name, msg_dict in errors:
                    chunk.errors.append(message.message_name_to_class_map[msg_name](**msg_dict))
            warnings = chunk_cache.pop('warnings', None)
            if warnings is not None:
                for msg_name, msg_dict in warnings:
                    chunk.warnings.append(message.message_name_to_class_map[msg_name](**msg_dict))
            for output_name, output in chunk_cache.items():
                setattr(chunk, output_name, output)
        self._cached_sessions.add(session)
        return True


    def _update_session_cache(self, update_session: Session):
        '''
        Update cache for a session.

        Notice that individual cache files contain output from multiple
        sessions in the rare case that the session hashes have identical
        starting sequences (`.hash_root`).  In that case, a cache file may be
        overwritten multiple times during a document build as additional
        sessions are executed.  While this is not maximally efficient from a
        file system perspective, it will be rare and it ensures that code
        output is preserved if at all possible in the event of an unexpected
        exit.
        '''
        if self.no_cache or update_session.status.prevent_caching:
            return
        hash_root_cache = {
            'codebraid_version': codebraid_version,
            'cache': {},
        }
        for session in self._session_hash_root_sets[update_session.hash_root]:
            if session.status.prevent_caching:
                continue
            session_code_chunks_cache = {}
            for index, chunk in enumerate(session.code_chunks):
                chunk_cache = {}
                if chunk.errors:
                    errors_cache = [(x.type, x.as_dict()) for x in chunk.errors if x.is_cacheable]
                    if errors_cache:
                        chunk_cache['errors'] = errors_cache
                if chunk.warnings:
                    warnings_cache = [(x.type, x.as_dict()) for x in chunk.warnings if x.is_cacheable]
                    if warnings_cache:
                        chunk_cache['warnings'] = warnings_cache
                for output_name in ('stdout_lines', 'stderr_lines', 'repl_lines', 'expr_lines', 'rich_output'):
                    output = getattr(chunk, output_name)
                    if output is not None:
                        chunk_cache[output_name] = output
                if chunk_cache:
                    # `str(index)` because keys for JSON cache are strings
                    session_code_chunks_cache[str(index)] = chunk_cache
            hash_root_cache['cache'][session.hash] = {
                'session_errors': [(x.type, x.as_dict()) for x in session.errors if x.is_cacheable],
                'session_warnings': [(x.type, x.as_dict()) for x in session.warnings if x.is_cacheable],
                'session_files': session.files,
                'code_chunks': session_code_chunks_cache,
            }
        hash_root_cache_path = self._cache_key_path / f'{update_session.hash_root}.zip'
        with zipfile.ZipFile(str(hash_root_cache_path), 'w', compression=zipfile.ZIP_DEFLATED) as zf:
            zf.writestr('cache.json', json.dumps(hash_root_cache))
        self._cached_sessions.add(update_session)


    def _update_cache_index(self):
        if self.no_cache:
            return
        used_cache_files = set()
        used_cache_files.add(f'{self.cache_key}_index.zip')
        for session in self._cached_sessions:
            used_cache_files.add(f'{session.hash_root}.zip')
            used_cache_files.update(session.files)
        if self._old_cache_index is not None:
            unused_cache_files = set(self._old_cache_index['files']) - used_cache_files
            for f in unused_cache_files:
                try:
                    (self._cache_key_path / f).unlink()
                except FileNotFoundError:
                    pass
        new_cache_index = {
            'codebraid_version': codebraid_version,
            'origins': self._origin_paths_for_cache_as_strings,
            'files': list(used_cache_files),
        }
        with zipfile.ZipFile(str(self._cache_index_path), 'w', compression=zipfile.ZIP_DEFLATED) as zf:
            zf.writestr('index.json', json.dumps(new_cache_index))


    def cleanup(self):
        try:
            self._cache_lock_path.unlink()
        except FileNotFoundError:
            pass
        if self.no_cache:
            for session in self._sessions.values():
                for f in session.files:
                    try:
                        (self._cache_key_path / f).unlink()
                    except FileNotFoundError:
                        pass
            try:
                self._cache_key_path.rmdir()
            except Exception:
                pass
            else:
                try:
                    self.cache_path.rmdir()
                except Exception:
                    pass
        atexit.unregister(self.cleanup)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2021-2022, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the BSD 3-Clause License:
# http://opensource.org/licenses/BSD-3-Clause
#


from __future__ import annotations


import asyncio
import collections
import datetime
import itertools
import json
import sys
import time
import textwrap
from typing import Callable, Deque, Dict, Iterable, List, Optional
from .terminal import stderr_term
from .code_chunks import CodeChunk, CodeKey
from .code_collections import CodeCollection, Session, Source
from . import util




class Progress(object):
    '''
    Centralized tracking of Codebraid progress, from parsing original document
    source files to executing code to converting.  This makes possible a
    terminal progress display with customizable levels of detail.  It also
    makes possible arbitrary tasks that use code output immediately after it
    becomes available.
    '''
    def __init__(self, *, only_code_output: Optional[str], no_execute: bool):
        self._only_code_output = only_code_output
        if only_code_output:
            self._only_code_output_first_chunk_cache: dict[tuple, str] = {}
        self._no_execute = no_execute

        self._last_error_counts: Dict[CodeKey, int] = {}
        self._error_count: int = 0
        self._last_warning_counts: Dict[CodeKey, int] = {}
        self._warning_count: int = 0

        self._session_total_chunks_count: int = 0
        self._session_exec_chunks_count: int = 0
        self._session_exec_completed_chunks_count: int = 0
        self._session_exec_last_completed_chunk_count: Dict[CodeKey, int] = collections.defaultdict(int)

        self._current_task: Optional[str] = None
        self._current_subtask: Optional[str] = None
        self._last_task: Optional[str] = None
        self._last_subtask: Optional[str] = None
        self._last_progress_time: float = time.monotonic()
        if self.term.isatty:
            ellipsis_sequence = ['   ', '.  ', '.. ', '...']
        else:
            ellipsis_sequence = ['   ']
        self._progress_ellipsis_deque: Deque = collections.deque(ellipsis_sequence)
        self._in_live_output: bool = False
        self._last_live_output: Optional[str] = None
        self._last_live_output_stream: Optional[str] = None
        self._live_output_backlog: List[str] = []


    term = stderr_term


    _first_textwrappers = util.KeyDefaultDict(
        lambda columns: textwrap.TextWrapper(width=columns, initial_indent='  * ', subsequent_indent=' '*4)
    )
    _subsequent_textwrappers = util.KeyDefaultDict(
        lambda columns: textwrap.TextWrapper(width=columns, initial_indent=' '*4, subsequent_indent=' '*4)
    )


    def register_code_collections(self, *, sessions: Iterable[Session], sources: Iterable[Source]):
        for code_collection in itertools.chain(sessions, sources):
            self._last_error_counts[code_collection.key] = code_collection.status.error_count
            self._error_count += code_collection.status.error_count
            self._last_warning_counts[code_collection.key] = code_collection.status.warning_count
            self._warning_count += code_collection.status.warning_count
        self._session_total_chunks_count = sum(len(s.code_chunks) for s in sessions)
        self._session_exec_chunks_count = self._session_total_chunks_count
        if self._only_code_output:
            code_collections_list = []
            placeholder_langs_dict = {}
            for code_collection in itertools.chain(sessions, sources):
                if isinstance(code_collection, Source) or self._no_execute:
                    will_execute = False
                else:
                    will_execute = code_collection.needs_exec
                code_collections_list.append({
                    'type': code_collection.type,
                    'lang': code_collection.lang or '',
                    'name': code_collection.name or '',
                    'origin_name': code_collection.origin_name or '',
                    'length': len(code_collection.code_chunks),
                    'will_execute': will_execute,
                })
                for placeholder_lang in code_collection.code_chunk_placeholder_langs:
                    placeholder_langs_dict[placeholder_lang] = code_collection.lang or ''
            data = {
                'message_type': 'index',
                'code_collections': code_collections_list,
                'placeholder_langs': placeholder_langs_dict
            }
            # All data I/O must be UTF-8, following Pandoc
            sys.stdout.buffer.write(json.dumps(data).encode('utf8'))
            sys.stdout.buffer.write(b'\n')
            sys.stdout.buffer.flush()


    def _code_messages_to_summary_list(self, code_collection: CodeCollection, *, msg_type: str, columns: int) -> List[str]:
        first_textwrapper = self._first_textwrappers[columns]
        subsequent_textwrapper = self._subsequent_textwrappers[columns]
        if msg_type == 'errors':
            fmt_msg_type = self.term.fmt_error
        elif msg_type == 'warnings':
            fmt_msg_type = self.term.fmt_warning
        else:
            raise ValueError
        summary_list = []
        for msg in getattr(code_collection, msg_type):
            if msg.is_refed:
                continue
            if summary_list:
                summary_list.append('')
            for line in first_textwrapper.wrap(f'"{code_collection.code_chunks[0].origin_name}":'):
                summary_list.append(fmt_msg_type(line))
            for line in msg.message:
                summary_list.extend(subsequent_textwrapper.wrap(line))
        for cc in code_collection.code_chunks:
            for msg in getattr(cc, msg_type):
                if msg.is_refed:
                    continue
                if summary_list:
                    summary_list.append('')
                for line in first_textwrapper.wrap(f'"{cc.origin_name}", line {cc.origin_start_line_number}:'):
                    summary_list.append(fmt_msg_type(line))
                for line in msg.message:
                    summary_list.extend(subsequent_textwrapper.wrap(line))
        return summary_list

    def _summarize_code_messages(self, code_collection: CodeCollection, *, columns: int) -> str:
        summary_list = []
        if code_collection.status.has_errors:
            summary_list.append(self.term.fmt_error('Errors:'))
            summary_list.extend(self._code_messages_to_summary_list(code_collection, msg_type='errors', columns=columns))
        if code_collection.status.has_warnings:
            summary_list.append(self.term.fmt_warning('Warnings:'))
            summary_list.extend(self._code_messages_to_summary_list(code_collection, msg_type='warnings', columns=columns))
        summary_list.append('')
        return '\n'.join(summary_list)


    def _update_progress(self):
        '''
        Update terminal progress summary, which includes current task, current
        error and warnings counts, and code execution progress.
        '''
        if self._only_code_output or self._current_task is None:
            return
        if self._last_live_output is not None and not self._last_live_output.endswith('\n'):
            return
        time_now = time.monotonic()
        if (not self._in_live_output and
                self._current_task == self._last_task and self._current_subtask == self._last_subtask and
                time_now - self._last_progress_time < 1.0):
            return
        error_status = f'Errors: {self._error_count}'
        if self._error_count > 0:
            error_status_fmted = self.term.fmt_error(error_status)
        else:
            error_status_fmted = self.term.fmt_ok(error_status)
        warning_status = f'Warnings: {self._warning_count}'
        if self._warning_count > 0:
            warning_status_fmted = self.term.fmt_warning(warning_status)
        else:
            warning_status_fmted = self.term.fmt_ok(warning_status)
        if self._current_task == self._last_task:
            self._progress_ellipsis_deque.rotate(-1)
        else:
            while self._progress_ellipsis_deque[0] != '   ':
                self._progress_ellipsis_deque.rotate(-1)
        ellipsis = self._progress_ellipsis_deque[0]
        if self._current_subtask is None:
            task_w_subtask = self._current_task
            if self._current_task != 'Complete':
                task_w_subtask_fmted = self._current_task
            elif self._error_count > 0:
                task_w_subtask_fmted = self.term.fmt_error(self._current_task)
            elif self._warning_count > 0:
                task_w_subtask_fmted = self.term.fmt_warning(self._current_task)
            else:
                task_w_subtask_fmted = self.term.fmt_ok(self._current_task)
        else:
            task_w_subtask = f'{self._current_task}: {self._current_subtask}'
            task_w_subtask_fmted = task_w_subtask
        general_status = f'{task_w_subtask}{ellipsis} {error_status}  {warning_status}'
        general_status_fmted = f'{task_w_subtask_fmted}{ellipsis} {error_status_fmted}  {warning_status_fmted}'
        if self._current_task == 'Exec' and self._current_subtask == 'run':
            exec_status = f'  code chunk {self._session_exec_completed_chunks_count}/{self._session_exec_chunks_count}'
            bar_width = self.term.columns() - len(general_status) - len(exec_status) - 3
            finished_ratio = self._session_exec_completed_chunks_count / self._session_exec_chunks_count
            finished = round(finished_ratio*bar_width)
            unfinished = bar_width - finished
            bar = f''' [{'#'*finished}{'.'*unfinished}]'''
            filler = ''
        else:
            exec_status = ''
            bar = ''
            filler = ' '*(self.term.columns() - len(general_status))
        if self.term.isatty:
            progress_text = f'\r{general_status_fmted}{exec_status}{bar}{filler}'
        else:
            progress_text = f'PROGRESS: {general_status_fmted}{exec_status}{bar}\n'
        self._last_task = self._current_task
        self._last_subtask = self._current_subtask
        self._last_progress_time = time_now
        print(progress_text, file=self.term.stream, end='', flush=True)


    async def ticktock(self):
        '''
        Update terminal progress summary at 1 second intervals.
        '''
        if self.term.isatty:
            while True:
                try:
                    await asyncio.sleep(1)
                    self._update_progress()
                except asyncio.CancelledError:
                    break


    def _print_live_heading(self, code_collection: CodeCollection, *, notification_type: str, title: str,
                            columns: int, flush: bool, clearline: bool,
                            content_sep: bool=True, chunk: Optional[CodeChunk]=None):
        output_list = []
        if clearline:
            output_list.append(self.term.clearline(columns))
        output_list.append(self.term.fmt_delim('='*columns))
        datetime_now = datetime.datetime.now()
        timestamp = f'[{datetime_now.hour:02d}:{datetime_now.minute:02d}:{datetime_now.second:02d}]'
        sep = ' '*(columns-len(notification_type)-2-len(title)-len(timestamp))
        output_list.append(self.term.fmt_notify(f'{notification_type}: {title}{sep}{timestamp}'))
        if chunk is None:
            chunk_progress = ''
            chunk_traceback = ''
        else:
            chunk_progress = f', code chunk {chunk.index+1}/{len(code_collection.code_chunks)}'
            chunk_traceback = f'\n"{chunk.origin_name}", line {chunk.origin_start_line_number}'
        output_list.append(self.term.fmt_notify(f'{code_collection.lang}, {code_collection.type} "{code_collection.name or "<default>"}"{chunk_progress}{chunk_traceback}'))
        if content_sep:
            output_list.append(self.term.fmt_delim('.'*columns))
        print('\n'.join(output_list), file=self.term.stream, flush=flush)

    def _print_live_closing(self, code_collection: CodeCollection, *, notification_type: str, title: str,
                            columns: int, flush: bool, clearline: bool, chunk: Optional[CodeChunk]=None):
        output_list = []
        if self._last_live_output is not None and not self._last_live_output.endswith('\n'):
            output_list.append('\n')
        if clearline:
            output_list.append(self.term.clearline(columns))
        output_list.append(self.term.fmt_delim('-'*columns))
        output_list.append('\n')
        print(''.join(output_list), file=self.term.stream, flush=flush)

    def _print_live_notification(self, code_collection: CodeCollection, *, notification_type: str, title: str,
                                 columns: Optional[int]=None, text: Optional[str]=None):
        if columns is None:
            columns = self.term.columns()
        self._print_live_heading(code_collection, notification_type=notification_type, title=title,
                                 columns=columns, flush=False, clearline=self.term.isatty,
                                 content_sep=text is not None)
        if text is not None:
            if text.endswith('\n'):
                print(text, file=self.term.stream, end='')
            else:
                print(text, file=self.term.stream)
        self._print_live_closing(code_collection, notification_type=notification_type, title=title,
                                 columns=columns, flush=True, clearline=False)

    def _print_live_output(self, output: str, *, stream: str, fmter: Optional[Callable[[str], str]]=None):
        if self._last_live_output is not None and not self._last_live_output.endswith('\n'):
            if stream == self._last_live_output_stream:
                if fmter is None:
                    print(output, file=self.term.stream, end='', flush=True)
                else:
                    print(fmter(output), file=self.term.stream, end='', flush=True)
                self._last_live_output = output
                self._last_live_output_stream = stream
            elif stream == 'stderr':
                if '\r' in output and ('\n' not in output or output.find('\r') < output.find('\n')):
                    print('\n', file=self.term.stream, end='')
                if fmter is None:
                    print(output, file=self.term.stream, end='', flush=True)
                else:
                    print(fmter(output), file=self.term.stream, end='', flush=True)
                self._last_live_output = output
                self._last_live_output_stream = stream
            else:
                self._live_output_backlog.append((output, stream, fmter))
            return
        if self.term.isatty:
            print(self.term.clearline(), file=self.term.stream, end='')
        if not self._live_output_backlog:
            if fmter is None:
                print(output, file=self.term.stream, end='', flush=True)
            else:
                print(fmter(output), file=self.term.stream, end='', flush=True)
            self._last_live_output = output
            self._last_live_output_stream = stream
            return
        new_backlog = []
        for backlog_output, backlog_stream, backlog_fmter in self._live_output_backlog:
            if self._last_live_output.endswith('\n') or backlog_stream == self._last_live_output_stream:
                if backlog_fmter is None:
                    print(backlog_output, file=self.term.stream, end='')
                else:
                    print(backlog_fmter(backlog_output), file=self.term.stream, end='')
                self._last_live_output = backlog_output
                self._last_live_output_stream = backlog_stream
            else:
                new_backlog.append((backlog_output, backlog_stream, backlog_fmter))
        self._live_output_backlog = new_backlog
        if self._last_live_output.endswith('\n'):
            if fmter is None:
                print(output, file=self.term.stream, end='')
            else:
                print(fmter(output), file=self.term.stream, end='')
            self._last_live_output = output
            self._last_live_output_stream = stream
        elif stream == 'stderr':
            if '\r' in output and ('\n' not in output or output.find('\r') < output.find('\n')):
                print('\n', file=self.term.stream, end='')
            if fmter is None:
                print(output, file=self.term.stream, end='')
            else:
                print(fmter(output), file=self.term.stream, end='')
            self._last_live_output = output
            self._last_live_output_stream = stream
        else:
            self._live_output_backlog.append((output, stream, fmter))
        self.term.stream.flush()


    def _print_code_output(self, code_collection: CodeCollection, *, chunk: CodeChunk, check_first_chunk=False):
        if isinstance(code_collection, Session) and code_collection.did_exec:
            exec_progress = f'{self._session_exec_completed_chunks_count}/{self._session_exec_chunks_count}'
        else:
            exec_progress = None
        data = {
            'message_type': 'output',
            'origin': chunk.origin_name or '',
            'code_collection': {
                'type': code_collection.type,
                'lang': chunk.options['lang'] or '',
                'name': code_collection.name or '',
                'origin_name': code_collection.origin_name or '',
            },
            'inline': chunk.inline,
            'number': f'{chunk.index+1}/{len(code_collection.code_chunks)}',
            'exec_progress': exec_progress,
            'attr_hash': chunk.attr_hash,
            'code_hash': chunk.code_hash,
            'output': chunk.only_code_output(self._only_code_output),
        }
        data_str = json.dumps(data)
        if check_first_chunk:
            old_data_str = self._only_code_output_first_chunk_cache.get(code_collection.key)
            if old_data_str == data_str:
                return
        elif chunk.index == 0:
            self._only_code_output_first_chunk_cache[code_collection.key] = data_str
        # All data I/O must be UTF-8, following Pandoc
        sys.stdout.buffer.write(data_str.encode('utf8'))
        sys.stdout.buffer.write(b'\n')
        sys.stdout.buffer.flush()


    def _update_message_count(self, code_collection: CodeCollection):
        self._error_count += code_collection.status.error_count - self._last_error_counts[code_collection.key]
        self._last_error_counts[code_collection.key] = code_collection.status.error_count
        self._warning_count += code_collection.status.warning_count - self._last_warning_counts[code_collection.key]
        self._last_warning_counts[code_collection.key] = code_collection.status.warning_count


    def parse_start(self):
        self._current_task = 'Parse'
        self._current_subtask = None
        self._update_progress()

    def parse_end(self):
        pass

    def process_start(self):
        self._current_task = 'Process'
        self._current_subtask = None
        self._update_progress()

    def process_end(self):
        pass

    def exec_start(self):
        self._current_task = 'Exec'
        self._current_subtask = None
        self._update_progress()

    def exec_end(self):
        pass

    def convert_start(self):
        self._current_task = 'Convert'
        self._current_subtask = None
        self._update_progress()

    def convert_end(self):
        pass

    def complete(self):
        self._current_task = 'Complete'
        self._update_progress()


    def session_load_cache(self, session: Session):
        self._update_message_count(session)
        self._session_exec_chunks_count -= len(session.code_chunks)
        if session.live_output:
            self._in_live_output = True
            self._print_live_notification(session, notification_type='SESSION', title='LOADED CACHE')
        self._update_progress()

    def session_errors_prevent_exec(self, session: Session):
        self._update_message_count(session)
        self._session_exec_chunks_count -= len(session.code_chunks)

    def session_exec_stage_start(self, session: Session, *, stage: str):
        self._current_subtask = stage
        if session.live_output:
            self._in_live_output = True
            self._print_live_notification(session, notification_type='SESSION', title=f'START {stage.upper()}')
        self._update_progress()

    def session_exec_stage_output(self, session: Session, *, output: str):
        if session.live_output:
            self._print_live_output(output, stream='stdout')

    def session_exec_stage_end(self, session: Session, *, stage: str):
        self._update_message_count(session)
        self._update_progress()


    def _finished(self, code_collection: CodeCollection):
        self._update_message_count(code_collection)
        if (code_collection.type == 'session' and code_collection.live_output and
                (code_collection.status.has_errors or code_collection.status.has_warnings)):
            columns = self.term.columns()
            self._print_live_notification(code_collection, notification_type=f'{code_collection.type.upper()}',
                                          title='SUMMARY', columns=columns,
                                          text=self._summarize_code_messages(code_collection, columns=columns))
        self._update_progress()
        self._in_live_output = False

    def source_finished(self, source: Source):
        # No need for conditional ._only_code_output processing here for
        # missed chunks, since unlike the code execution case, source chunks
        # always update their status when complete.
        if self._only_code_output:
            self._print_code_output(source, chunk=source.code_chunks[0], check_first_chunk=True)
        self._finished(source)

    def session_finished(self, session: Session):
        if session.did_exec:
            # If session exec is interrupted by errors, code chunk count may
            # be off because per-chunk progress wasn't registered
            delta = len(session.code_chunks) - self._session_exec_last_completed_chunk_count[session.key]
            self._session_exec_completed_chunks_count += delta
            if self._only_code_output:
                for index in range(self._session_exec_last_completed_chunk_count[session.key], len(session.code_chunks)):
                    self._print_code_output(session, chunk=session.code_chunks[index])
                self._print_code_output(session, chunk=session.code_chunks[0], check_first_chunk=True)
        self._finished(session)

    def source_chunk_complete(self, source: Source, *, chunk: CodeChunk):
        if self._only_code_output:
            self._print_code_output(source, chunk=chunk)

    def session_chunk_complete_no_exec(self, session: Session, *, chunk: CodeChunk):
        if self._only_code_output:
            self._print_code_output(session, chunk=chunk)

    def session_chunk_start_exec(self, session: Session, *, chunk: CodeChunk):
        if session.live_output:
            self._print_live_heading(session, chunk=chunk, notification_type='CODE CHUNK', title='LIVE OUTPUT',
                                     columns=self.term.columns(), flush=True, clearline=self.term.isatty)
        self._update_progress()

    def session_chunk_end_exec(self, session: Session, *, chunk: CodeChunk):
        self._update_message_count(session)
        old_session_exec_last_completed_chunk_count = self._session_exec_last_completed_chunk_count[session.key]
        self._session_exec_completed_chunks_count += chunk.index + 1 - old_session_exec_last_completed_chunk_count
        self._session_exec_last_completed_chunk_count[session.key] = chunk.index + 1
        if self._only_code_output:
            for index in range(old_session_exec_last_completed_chunk_count, chunk.index+1):
                self._print_code_output(session, chunk=session.code_chunks[index])
        if session.live_output:
            self._print_live_closing(session, chunk=chunk, notification_type='CODE CHUNK', title='LIVE OUTPUT',
                                     columns=self.term.columns(), flush=True, clearline=self.term.isatty)
            self._last_live_output = None
            self._last_live_output_stream = None
        self._update_progress()


    def session_chunk_stdout(self, session: Session, *, chunk: Optional[CodeChunk], output: str):
        if session.live_output:
            self._print_live_output(output, stream='stdout')

    def session_chunk_stderr(self, session: Session, *, chunk: Optional[CodeChunk], output: str):
        if session.live_output:
            self._print_live_output(output, stream='stderr', fmter=self.term.fmt_error)

    def session_chunk_rich_output_text(self, session: Session, *, chunk: CodeChunk, output: str):
        if session.live_output:
            output_lines = util.splitlines_lf(output)
            columns = self.term.columns()
            output_list = ['RICH OUTPUT:']
            output_list.extend(self._first_textwrappers[columns].wrap(output_lines[0]))
            for line in output_lines[1:]:
                output_list.extend(self._subsequent_textwrappers[columns].wrap(line))
            output_list.append('')
            self._print_live_output('\n'.join(output_list), stream='rich_output', fmter=self.term.fmt_notify)

    def session_chunk_rich_output_files(self, session: Session, *, chunk: CodeChunk, files: Iterable[str]):
        if session.live_output:
            output_list = ['RICH OUTPUT FILES:\n']
            for file in files:
                output_list.append(f'  * {file}\n')
            self._print_live_output(''.join(output_list), stream='rich_output_files', fmter=self.term.fmt_notify)

    def session_chunk_expr(self, session: Session, *, chunk: CodeChunk, output: str):
        if session.live_output:
            output_list = ['EXPR:']
            output_list.extend(self._first_textwrappers[self.term.columns()].wrap(output))
            self._print_live_output('\n'.join(output_list), stream='expr', fmter=self.term.fmt_notify)

    def session_chunk_repl(self, session: Session, *, chunk: CodeChunk, output: str):
        if session.live_output:
            self._print_live_output(output, stream='repl')



# -*- coding: utf-8 -*-
#
# Copyright (c) 2021-2022, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the BSD 3-Clause License:
# http://opensource.org/licenses/BSD-3-Clause
#


from __future__ import annotations


from typing import Dict, List, Optional, Type, Union
from . import util


message_name_to_class_map: Dict[str, 'BaseMessage'] = {}


class MetaMessage(type):
    '''
    Metaclass for message class.  Check class attributes and create a mapping
    from class names (strings) to classes that is used in caching.
    '''
    def __new__(cls, name, parents, attr_dict):
        attr_dict['type'] = name
        display_name_list = []
        for c in name:
            if ord('A') <= ord(c) <= ord('Z') and display_name_list:
                display_name_list.append(' ')
            display_name_list.append(c)
        attr_dict['display_name'] = ''.join(display_name_list)
        attr_dict['is_stderr'] = 'Stderr' in name
        attr_dict['is_ref'] = name.endswith('Ref')
        new_class = super().__new__(cls, name, parents, attr_dict)
        if hasattr(new_class, 'category'):
            if not isinstance(new_class.category, str):
                raise TypeError
            for attr in cls._checked_attrs:
                if not hasattr(new_class, attr) or not isinstance(getattr(new_class, attr), bool):
                    raise TypeError
            if 'Error' in name:
                if not new_class.prevent_exec and 'CanExec' not in name:
                    raise ValueError
            elif 'Warning' in name:
                if new_class.prevent_caching or new_class.prevent_exec:
                    raise ValueError
            else:
                raise ValueError
            if new_class.prevent_caching and new_class.is_cacheable:
                raise ValueError
            if not new_class.is_cacheable:
                if 'Source' not in name and 'SysConfig' not in name:
                    raise ValueError
        message_name_to_class_map[name] = new_class
        return new_class
    _checked_attrs = ('prevent_caching', 'is_cacheable', 'prevent_exec')


class BaseMessage(object, metaclass=MetaMessage):
    '''
    Base class for all error and warning messages.

    Cannot be instantiated.
    '''
    def __init__(self, message: Optional[Union[str, List[str]]], is_refed: bool=False):
        if not hasattr(self, 'category'):
            raise NotImplementedError
        if isinstance(message, str):
            message = util.splitlines_lf(message)
        self.message: Optional[List[str]] = message
        self.is_refed: bool = is_refed

    # General category to which message belongs.
    category: str
    # Whether this particular message is cacheable when caching is not
    # prevented.  Messages that are not cacheable must be regenerated each
    # time during document processing.  Messages that are cacheable typically
    # relate to (attempted) code execution.
    is_cacheable: bool
    # Whether this message prevents all caching of messages and code output.
    # For simplicity, errors that are detected before attempting code
    # execution prevent caching, since it is typically trivial to regenerate
    # such errors each time during document processing.  Errors that are due
    # to unmonitored factors like system configuration also prevent caching,
    # since the only way to determine if they have been resolved is to try
    # executing code again.
    prevent_caching: bool
    # Whether this message is severe enough to completely prevent any
    # subsequent code execution.
    prevent_exec: bool
    # Whether this message is simply a copy of information present in stderr.
    is_stderr: bool
    # Whether this message is a reference to another message.  For example, a
    # code chunk can have a stderr message that references a session-level
    # stderr message.
    is_ref: bool
    # Name of class as a string.  Used in caching.
    type: str
    # Name of class as string with spaces inserted.
    display_name: str

    def as_dict(self):
        return {
            'message': self.message,
            'is_refed': self.is_refed,
        }


class BaseError(BaseMessage):
    '''
    Base class for all error messages.
    '''

class SourceError(BaseError):
    '''
    Error in source, such as a typo or invalid value, that might affect code
    execution.
    '''
    category = 'source'
    is_cacheable = False
    prevent_caching = True
    prevent_exec = True

class CanExecSourceError(SourceError):
    '''
    Error in source, such as a typo or invalid value, that cannot affect code
    execution.
    '''
    prevent_caching = False
    prevent_exec = False

class RuntimeSourceError(BaseError):
    '''
    Error in source that can only be detected at runtime.

    For example, a code chunk with `complete=true` that is actually not
    complete.
    '''
    category = 'runtime_source'
    is_cacheable = True
    prevent_caching = False
    prevent_exec = True

class SysConfigError(BaseError):
    '''
    Error in system configuration.

    For example, attempting to use a Jupyter kernel without having
    `jupyter_client` installed, or trying to run a program that is not in
    PATH.

    This error is not cached since the system configuration is not monitored,
    so it might change by the next attempt and allow success.
    '''
    category: str = 'sys_config'
    is_cacheable = False
    prevent_caching = True
    prevent_exec = True

class ExecError(BaseError):
    '''
    Error during the compile, pre-run, run, or post-run stage of code
    execution.

    Typically, the message will consist of stderr for the run stage, and the
    combined stdout and stderr for other stages.
    '''
    def __init__(self, message: Optional[Union[str, List[str]]], *,
                 is_refed: bool=False, exit_code: Optional[int]=None):
        super().__init__(message, is_refed=is_refed)
        self.exit_code: Optional[int] = exit_code
    is_cacheable = True
    prevent_caching = False
    prevent_exec = True
    def as_dict(self):
        error_as_dict = super().as_dict()
        error_as_dict['exit_code'] = self.exit_code
        return error_as_dict

class CompileError(ExecError):
    category: str = 'compile'

class PreRunError(ExecError):
    category: str = 'pre-run'

class RunError(ExecError):
    category: str = 'run'

class StderrRunError(RunError):
    # Run error that is also present in stderr
    pass

class StderrRunErrorRef(RunError):
    # Reference from code chunk to session run error that is also present in
    # stderr
    pass

class RunConfigError(ExecError):
    category: str = 'run_config'

class PostRunError(ExecError):
    category: str = 'post-run'

class DecodeError(BaseError):
    '''
    Error in decoding code output.
    '''
    category: str = 'encoding'
    is_cacheable = True
    prevent_caching = False
    prevent_exec = True


class BaseWarning(BaseMessage):
    '''
    Base class for all warning messages.

    Warnings cannot be sufficiently severe to impact caching or code
    execution; only errors can do that.
    '''
    prevent_caching = False
    prevent_exec = False

class SourceWarning(BaseWarning):
    '''
    Warning about source.
    '''
    category: str = 'source'
    is_cacheable = False

class ExecWarning(BaseWarning):
    '''
    Warning during the compile, pre-run, run, or post-run stage of code
    execution.

    Typically, the message will consist of stderr for the run stage, and the
    combined stdout and stderr for other stages.
    '''

class RunWarning(ExecWarning):
    category: str = 'run'
    is_cacheable = True

class StderrRunWarning(RunWarning):
    # Run warning that is also present in stderr
    pass

class StderrRunWarningRef(RunWarning):
    # Reference from code chunk to session run warning that is also present in
    # stderr
    pass

class CodeStatus(object):
    '''
    Status of code belonging to a Session or Source.
    '''
    def __init__(self):
        # Attributes summarize the overall attributes of
        self.prevent_caching: bool = False
        self.prevent_exec: bool = False
        self.error_count: int = 0
        self.warning_count: int = 0
        self.has_stderr: bool = False
        self.has_non_stderr: bool = False

    @property
    def has_errors(self) -> bool:
        return self.error_count > 0

    @property
    def has_warnings(self) -> bool:
        return self.warning_count > 0


class MessageList(list):
    '''
    Base class for list of Message objects that also tracks some properties of
    the collection.

    Cannot be instantiated.
    '''
    def __init__(self, *, status=None):
        if not hasattr(self, '_contents_base_class'):
            raise NotImplementedError
        # Attributes summarize the attributes of objects in list
        self.prevent_caching: bool = False
        self.prevent_exec: bool = False
        self.has_stderr: bool = False
        self.has_non_stderr: bool = False
        self._refed_list_ids: set[int] = set()
        self._list_id_to_msg: dict[int, BaseMessage] = {}

        self._status: Optional[CodeStatus] = status
        super().__init__()

    _contents_base_class: Union[Type[BaseError], Type[BaseWarning]]

    def append(self, msg: BaseMessage):
        # Prevent accidental mixing of error and warning objects
        if not isinstance(msg, self._contents_base_class):
            raise TypeError
        # Message list ids should be unique.  It may be worth adding a check
        # for duplicates at some point, depending on the final implementation
        # of stderr sync, etc.
        self._list_id_to_msg[id(msg.message)] = msg
        if not self.prevent_caching:
            self.prevent_caching = msg.prevent_caching
        if not self.prevent_exec:
            self.prevent_exec = msg.prevent_exec
        if not self.has_stderr:
            self.has_stderr = msg.is_stderr
        if not self.has_non_stderr:
            self.has_non_stderr = not msg.is_stderr
        if self._status is not None:
            if not self._status.prevent_caching:
                self._status.prevent_caching = msg.prevent_caching
            if not self._status.prevent_exec:
                self._status.prevent_exec = msg.prevent_exec
            if not self._status.has_stderr:
                self._status.has_stderr = msg.is_stderr
            if not self._status.has_non_stderr:
                self._status.has_non_stderr = not msg.is_stderr
        super().append(msg)

    def extend(self, msgs: List[BaseMessage]):
        for msg in msgs:
            self.append(msg)

    def register_status(self, status: CodeStatus):
        '''
        Update `status` if there are already messages, and set `_status` for
        future updating.
        '''
        if self._status is not None:
            raise RuntimeError
        if self:
            if self._contents_base_class is BaseError:
                for msg in self:
                    if not msg.is_ref:
                        status.error_count += 1
            elif self._contents_base_class is BaseWarning:
                for msg in self:
                    if not msg.is_ref:
                        status.warning_count += 1
            else:
                raise TypeError
            if not status.prevent_caching:
                status.prevent_caching = self.prevent_caching
            if not status.prevent_exec:
                status.prevent_exec = self.prevent_exec
        self._status = status

    def update_refed(self, message_list):
        self._list_id_to_msg[id(message_list)].is_refed = True

    def has_ref(self, message_or_message_list: BaseMessage | list[str]):
        if isinstance(message_or_message_list, list):
            return id(message_or_message_list) in self._list_id_to_msg
        if isinstance(message_or_message_list, BaseMessage):
            return id(message_or_message_list.message) in self._list_id_to_msg
        raise TypeError

class ErrorMessageList(MessageList):
    _contents_base_class = BaseError
    def append(self, msg: BaseError):
        if self._status is not None and not msg.is_ref:
            self._status.error_count += 1
        super().append(msg)

class WarningMessageList(MessageList):
    _contents_base_class = BaseWarning
    def append(self, msg: BaseWarning):
        if self._status is not None and not msg.is_ref:
            self._status.warning_count += 1
        super().append(msg)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2018, Geoffrey M. Poore
# All rights reserved.
#
# Licensed under the BSD 3-Clause License:
# http://opensource.org/licenses/BSD-3-Clause
#


from .version import __version__, __version_info__
